import smtplib
from email.message import EmailMessage
from typing import List, Optional, Tuple

from r2k.cli import logger

from .config import config
from .unicode import strip_common_unicode_chars


def build_basic_message(title: str) -> EmailMessage:
    msg = EmailMessage()
    msg["Subject"] = title
    msg["To"] = config.send_to
    msg["From"] = config.send_from
    return msg


def send_email_message(server: smtplib.SMTP, msg: EmailMessage) -> bool:
    """Send a single EmailMessage"""
    logger.debug("Sending the email...")
    try:
        server.send_message(msg)
        return True
    except smtplib.SMTPException as e:
        logger.error(f"Caught an exception while trying to send an email.\nError: {e}")
        return False


def send_email_messages(msgs: List[EmailMessage]) -> int:
    """Send an email"""
    messages_sent = 0
    try:
        logger.debug("Connecting to SMTP...")
        with smtplib.SMTP_SSL("smtp.gmail.com", 465) as server:
            server.ehlo()
            logger.debug("Logging into the SMTP server...")
            server.login(config.send_from, config.password)
            for msg in msgs:
                if send_email_message(server, msg):
                    messages_sent += 1
                    logger.notice("Email sent successfully!\n")
    except smtplib.SMTPException as e:
        logger.error(f"Caught an exception while trying to send an email.\nError: {e}")
    return messages_sent


def set_content(msg: EmailMessage, title: str, url: Optional[str], attachment_path: Optional[str]) -> None:
    """Either set the text content of the email message, or attach an attachment, based on the current parser"""
    if attachment_path:
        # We are marking the attachment as HTML, although it's an epub, because kindle doesn't official accept
        # EPUB files in emails, but unofficial it will convert the file with kindlegen and it'll work fine
        # Reference: https://www.amazon.com/gp/sendtokindle/email
        filename = f"{title}.html"
        logger.debug(f"Setting attachment for {title}")
        with open(attachment_path, "rb") as f:
            msg.add_attachment(
                f.read(), maintype="text", subtype=f'html; charset=utf-8; name="{filename}"', filename=filename,
            )
    elif url:
        logger.debug(f"Setting email content to {url}")
        msg.set_content(url)


def create_email_message(title: str, url: Optional[str], attachment_path: Optional[str]) -> EmailMessage:
    title = strip_common_unicode_chars(title)
    msg = build_basic_message(title)
    set_content(msg, title, url, attachment_path)
    return msg


def send_epub(title: str, epub_path: str) -> int:
    msg = create_email_message(title, None, epub_path)
    return send_email_messages([msg])


def send_urls(articles: List[Tuple[str, str]]) -> int:
    """Send a list of URLs to Kindle (via pushtokindle)"""
    msgs = []
    for title, url in articles:
        msgs.append(create_email_message(title, url, None))
    return send_email_messages(msgs)



import click

from r2k.cli import cli_utils, logger
from r2k.config import config


@click.command("remove")
@cli_utils.config_path_option()
@click.option("-t", "--title", type=str, required=True, help="The title of the feed")
def feed_remove(title: str) -> None:
    """Remove a feed from the config."""
    found = config.feeds.pop(title, False)
    config.save()

    if found:
        logger.info(f"Successfully removed `{title}` from the list of feeds")
    else:
        logger.info(f"Could not find `{title}` in the list of feeds")



from __future__ import annotations

from types import TracebackType
from typing import Optional, Type

from readability import Document
from requests import get

from r2k.constants import HTML_HEADERS

from .base_parser import ParserBase


class ReadabilityParser(ParserBase):
    """
    Parser that uses the readability module
    """

    def __enter__(self) -> ReadabilityParser:
        """Nothing to do here"""
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> bool:
        """Nothing to do here"""
        pass

    def parse(self, url: str) -> dict:
        """Download the article and parse it"""
        r = get(url, headers=HTML_HEADERS)
        doc = Document(r.text, url=url)
        return {"content": doc.summary(html_partial=True)}



from __future__ import annotations

from abc import ABC, abstractmethod
from types import TracebackType
from typing import Optional, Type


class ParserBase(ABC):
    """
    Base class for parsers
    """

    @abstractmethod
    def __enter__(self) -> ParserBase:
        """Context manager __enter__"""
        pass

    @abstractmethod
    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> bool:
        """Context manager __exit__"""
        pass

    @abstractmethod
    def parse(self, url: str) -> dict:
        """Main function used to parse the article from a URL"""
        pass



from dataclasses import Field, asdict, dataclass, field, fields
from typing import Any, List, Optional

import yaml

from .cli import logger
from .constants import Parser


@dataclass
class Config:
    """Global configuration for the project"""

    feeds: dict
    password: str
    kindle_address: str
    send_from: str
    send_to: str = field(init=False, default="")
    parser: Parser = Parser.READABILITY

    # Internal properties not accessible outside the class
    _path: str = field(init=False, repr=False)
    _loaded: bool = field(init=False, repr=False, default=False)

    def __post_load__(self) -> None:
        """Tasks to perform after loading the config from the YAML"""
        self._loaded = True
        if self.parser == Parser.PUSH_TO_KINDLE:
            kindle_address = self.kindle_address.split("@")[0]
            self.send_to = f"{kindle_address}@pushtokindle.com"
        else:
            self.send_to = self.kindle_address

    def load(self, path: str) -> None:
        """Load configurations from a YAML file"""
        self._path = path

        logger.debug(f"Loading config from {self._path}")
        with open(self._path) as f:
            file_config = yaml.safe_load(f)

        self.__dict__.update(file_config)
        self.__post_load__()

    def __setattr__(self, key: str, value: Any) -> None:
        """Override in order to allow dumping changes to file"""
        super().__setattr__(key, value)
        self.save()

    def save(self, path: Optional[str] = None) -> None:
        """Dump the contents of the internal dict to file"""
        if not self._loaded:
            return

        if not self._path:
            if path:
                self._path = path
            else:
                raise FileNotFoundError("Path not set in Config. Need to run config.load before running config.save")

        with open(self._path, "w") as f:
            yaml.safe_dump(self.as_dict(), f, default_flow_style=False)

    def as_dict(self) -> dict:
        """Return the underlying dict"""
        return asdict(self)

    @classmethod
    def fields(cls) -> List[str]:
        """Return a tuple with all the publicly exposed fields of the Config class"""
        # Ignoring mypy typing validation here, as for some reason it assumes that f.type is always Field, but it isn't
        return [f.name for f in fields(cls) if isinstance(f.type, Field) and f.type.repr]  # type: ignore


config = Config(feeds={}, kindle_address="", password="", send_from="")



import os
from typing import Any, Callable, Union

import click

from r2k.constants import CONFIG_ENV_VAR, DEFAULT_CONFIG_PATH


def config_path_option(exists: bool = True) -> Callable:
    """
    A convenience decorator

    As we'll probably want to be able to specify the config in pretty much every console command
    """

    def decorator(f: Callable) -> Callable:
        option = click.option(
            "-p",
            "--path",
            default=DEFAULT_CONFIG_PATH,
            envvar=CONFIG_ENV_VAR,
            show_default=True,
            type=str,
            # When `expose_value` is set to False, the option is not passed to the click Commands.
            # In our case, the config path should always exist, unless we're in `r2k config init`.
            # As we're loading the config using the `load_config` callback, there's no need to pass the option on.
            expose_value=not exists,
            callback=load_config,
            help="Path to the configuration file",
        )
        return option(f)

    return decorator


def load_config(ctx: click.Context, param: Union[click.Option, click.Parameter], value: Any) -> Any:
    """Load the config immediately after getting the path"""
    if not value or ctx.resilient_parsing:
        return
    from r2k.config import config

    if param.expose_value:
        if is_verbose():
            click.echo("Not loading config as we're in `r2k config init`")
        return value
    else:
        if os.path.exists(value):
            config.load(value)
        else:
            click.secho(
                f"Could not locate a configuration in the specified path {value}\n"
                f"If you want to create a new configuration, run `r2k init`.\n"
                f"If you already have a configuration file, pass `-p PATH` to any of the CLI commands,\n"
                f"or pass it as as an env var to `{CONFIG_ENV_VAR}`",
                fg="red",
                bold=True,
            )
            ctx.exit(1)


def force_option(help_text: str) -> Callable:
    """
    A convenience decorator

    For various --force click options, that allows one to provide a custom help text
    """

    def decorator(f: Callable) -> Callable:
        option = click.option(
            "--force", default=False, show_default=True, is_flag=True, type=click.types.BOOL, help=help_text,
        )
        return option(f)

    return decorator


def get_dummy_context() -> click.Context:
    """Return a dummy click context to allow using eg click.echo when not running an actual click command"""
    ctx = click.Context(click.Command("dummy"))
    ctx.params["verbose"] = True
    return ctx


def get_global_context() -> click.Context:
    """Return the top-level global Click context"""
    ctx = click.get_current_context(silent=True) or get_dummy_context()

    while ctx.parent:
        ctx = ctx.parent
    return ctx


def is_verbose() -> bool:
    """Return True if --verbose was passed to the top-level click command"""
    return get_global_context().params.get("verbose", False)


def no_ansi() -> bool:
    """Return True if --no-ansi was passed to the top-level click command"""
    return get_global_context().params.get("no_ansi", False)



import sys
from typing import Optional
from xml.etree import ElementTree

import click
import yaml

from r2k.cli import cli_utils, logger
from r2k.config import config
from r2k.unicode import strip_common_unicode_chars


@click.command("import")
@cli_utils.config_path_option()
@click.argument("opml-path", required=True, type=click.types.Path(exists=True))
@cli_utils.force_option("If set will update existing feeds")
def feed_import(opml_path: str, force: bool) -> None:
    """Import feeds from an OPML file."""
    feeds = convert_opml_to_dict(opml_path)

    validate_conflicts(feeds, force)

    config.feeds.update(feeds)
    config.save()

    logger.info("Successfully imported the feeds!")


def validate_conflicts(feeds: dict, force: bool) -> None:
    """Error out if the force flag was not passed and there are conflicts between new and existing feeds"""
    new_feeds = set(feeds.keys())
    old_feeds = set(config.feeds.keys())
    conflicts = new_feeds & old_feeds
    if conflicts:
        conflicts_str = yaml.safe_dump(sorted(list(conflicts)))  # Just here to like nicer in the output
        if force:
            logger.confirm(f"Going to overwrite the following feeds:\n{conflicts_str}")
        else:
            logger.error(
                f"Found the following existing feeds:\n{conflicts_str}\n"
                f"Pass the --force flag if you'd like to overwrite them."
            )
            sys.exit(1)


def convert_opml_to_dict(path: str) -> dict:
    """Convert an OPML file to a dictionary, for easier storing in the configuration YAML"""
    tree = parse_ompl(path)
    feeds = {}
    for node in tree.findall(".//outline"):
        if title := get_title(node):
            if url := get_url(node, title):
                if is_rss(node, title):
                    feeds[title] = {"url": url}
    return feeds


def get_title(node: ElementTree.Element) -> Optional[str]:
    """Retrieve the feed's title from the XML element"""
    # The `title` and `text` are usually identical
    title = node.attrib.get("title", node.attrib.get("text"))
    if title:
        title = strip_common_unicode_chars(title)
    else:
        logger.debug("Could not find title for RSS feed")
    return title


def get_url(node: ElementTree.Element, title: str) -> Optional[str]:
    """Retrieve the feed's URL from the XML element"""
    url = node.attrib.get("xmlUrl")
    if not url:
        logger.debug(f"Could not find URL for `{title}`")
    return url


def is_rss(node: ElementTree.Element, title: str) -> bool:
    """Return true if the element is indeed an RSS feed"""
    rss_type = node.attrib.get("type")
    if rss_type != "rss":
        logger.debug(f"Unknown type for `{title}`: {rss_type}")
        return False
    return True


def parse_ompl(path: str) -> ElementTree.ElementTree:
    """Parse the OPML file and error out if it's in the wrong format"""
    with open(path) as f:
        try:
            return ElementTree.parse(f)
        except ElementTree.ParseError:
            logger.error(f"Could not parse `{path}`.\nIt's probably not a proper OPML file.")
            sys.exit(1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['r2k',
 'r2k.cli',
 'r2k.cli.config',
 'r2k.cli.feed',
 'r2k.cli.kindle',
 'r2k.ebook']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.0,<5.0.0',
 'click>=7.1.1,<8.0.0',
 'feedparser>=5.2.1,<6.0.0',
 'orjson>=2.6.6,<3.0.0',
 'pick>=0.6.7,<0.7.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'readability-lxml>=0.7.1,<0.8.0']

extras_require = \
{'docker': ['docker>=4.2.0,<5.0.0']}

entry_points = \
{'console_scripts': ['r2k = r2k.cli:main']}

setup_kwargs = {
    'name': 'r2k',
    'version': '0.3.7',
    'description': 'A tool that lets you periodically send articles received from an RSS feed to your Kindle',
    'long_description': '# R2K (RSS to Kindle)\n\nA tool that lets you periodically send articles received from an RSS feed to your Kindle.\n`r2k` allows you to track your favorite RSS/Atom feeds, and send cleaned versions of them to your\nKindle.\n\nIt currently does this by sending you email messages through the \n[PushToKindle](http://pushtokindle.com/) service. The service cleans up the articles, removing any\nads and fluff, and you\'re left with only the text and images.\n\nFuture releases are likely to include alternatives to using PushToKindle. The first such plugin\nwill use a dockerized version of the [Mercury-Parser](https://github.com/postlight/mercury-parser).  \n\n[Installation](#installation)\n[Usage](#usage)\n\n## Installation\nIf you are using Poetry, add `r2k` to your `pyproject.toml` file by using:\n```bash\npoetry add r2k\n```\n\nIf not, you can use `pip`:\n\n```\npip install r2k\n```\n\n## Usage\n\n### Preparations\n\n#### If you\'re using PushToKindle (PTK)\n\nThe free version of PTK only allows for 25 articles to be sent using their service. After this,\nyou\'ll have to become their "sustainer" (for as low as 1$/month) on Patreon \n[here](https://www.patreon.com/bePatron?c=1946606).\n\nBefore using `r2k` with PTK you need to:\n\n1. Know your kindle email address (find it [here](https://www.amazon.com/mycd), under \n"Preferences" -> "Personal Document Settings").\n\n2. Add `kindle@fivefilters.org` to the list of approved email senders (in the same place in \nAmazon\'s settings).\n\n### Set up your configuration file\n\nMost of what `r2k` does involves the configuration file, in which the feeds you\'re subscribed to\nare kept, as well as some other data.\n\nAfter installation, run:\n\n```bash\nr2k config init [-p CONFIG_PATH] \n```\n\nThe default location for the config YAML file is in `~/.r2k/config.yml`.\n\nDuring the init you\'ll be asked several questions (like your kindle email address).\n\nTo see your configuration run:\n\n```bash\nr2k config show [-p CONFIG_PATH]\n```\n\n### Add some RSS subscriptions\n\n#### Using an OPML file\n\nThe [OPML](https://en.wikipedia.org/wiki/OPML) format is widely used in the RSS/Atom world \n(as well as in podcasting and other areas) to represent a collection of feeds. You can export your \nexisting subscriptions from most feed readers into an OPML file.\n\nTo load all of your subscriptions in one move run:\n\n```bash\nr2k feed import PATH_TO_OPML_FILE\n```\n\n#### Manually adding feeds\n\nIf you don\'t have an OPML export, or just want to add a single feed you can run:\n\n```bash\nr2k feed add -t FEED_TITLE -u FEED_URL\n```\n\nIf the `FEED_URL` is a proper RSS feed (i.e. an actual XML feed URL), it will be added as is.\nIf the `FEED_URL` is a regular URL, `r2k` will attempt to find the RSS feed by analyzing the page\nsource. In the case of multiple candidates (e.g. WordPress content feed and comment feed), you will\nbe presented with a list of choices.\n\n### Send updates to your Kindle\n\nRight now the "periodical" part of `r2k` is not yet operational. In order to send updates to your\nKindle you\'ll have to run:\n\n```bash\nr2k kindle send [-f FEED_TITLE]\n```\n\nIf you don\'t pass the `-f/--feed-title` option, updates will be sent for all of your subscriptions.\n\nThe first time that `kindle send` is run for any feed, you will be presented with a list of all the \navailable articles in the feed (note that RSS feeds usually only keep a subset of the most recent\nentries), and will be asked to choose the last one you\'ve already read. This is to avoiding sending\nyou any article you\'ve already consumed.\n ',
    'author': 'Pavel Brodsky',
    'author_email': 'mcouthon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mcouthon/r2k',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
